"""

@author: agleim
"""

import numpy as np
from sklearn.utils.validation import check_array
from sklearn.base import BaseEstimator, TransformerMixin


__all__ = ["ChangeDetect"]


class ChangeDetect:

    pass
